@@ -51,7 +51,7 @@ typedef struct {
 	SoupEncoding          read_encoding;
 	GByteArray           *read_meta_buf;
 	SoupMessageBody      *read_body;
-	guint                 read_length;
+	goffset               read_length;
 
 	SoupMessageIOState    write_state;
 	SoupEncoding          write_encoding;
@@ -59,9 +59,11 @@ typedef struct {
 	SoupMessageBody      *write_body;
 	SoupBuffer           *write_chunk;
 	gsize                 write_body_offset;
-	guint                 written;
+	goffset               write_length;
+	goffset               written;
 
 	guint read_tag, write_tag, err_tag;
+	GSource *unpause_source;
 
 	SoupMessageGetHeadersFn   get_headers_cb;
 	SoupMessageParseHeadersFn parse_headers_cb;
@@ -139,6 +141,11 @@ soup_message_io_stop (SoupMessage *msg)
 		io->err_tag = 0;
 	}
 
+	if (io->unpause_source) {
+		g_source_destroy (io->unpause_source);
+		io->unpause_source = NULL;
+	}
+
 	if (io->read_state < SOUP_MESSAGE_IO_STATE_FINISHING)
 		soup_socket_disconnect (io->sock);
 	else if (io->conn) {
@@ -254,8 +261,7 @@ read_metadata (SoupMessage *msg, const char *boundary)
 /* Reads as much message body data as is available on io->sock (but no
  * further than the end of the current message body or chunk). On a
  * successful read, emits "got_chunk" (possibly multiple times), and
- * if %SOUP_MESSAGE_OVERWRITE_CHUNKS wasn't set, appends the chunk
- * to io->read_body.
+ * (unless told not to) appends the chunk to io->read_body.
  *
  * See the note at read_metadata() for an explanation of the return
  * value.
@@ -299,8 +305,7 @@ read_body_chunk (SoupMessage *msg)
 
 		if (status == SOUP_SOCKET_OK && nread) {
 			buffer->length = nread;
-			if (!(priv->msg_flags & SOUP_MESSAGE_OVERWRITE_CHUNKS))
-				soup_message_body_append_buffer (io->read_body, buffer);
+			soup_message_body_got_chunk (io->read_body, buffer);
 
 			io->read_length -= nread;
 
@@ -337,13 +342,15 @@ read_body_chunk (SoupMessage *msg)
  * read_metadata() for an explanation of the return value.
  */
 static gboolean
-write_data (SoupMessage *msg, const char *data, guint len)
+write_data (SoupMessage *msg, const char *data, guint len, gboolean body)
 {
 	SoupMessagePrivate *priv = SOUP_MESSAGE_GET_PRIVATE (msg);
 	SoupMessageIOData *io = priv->io_data;
 	SoupSocketIOStatus status;
 	gsize nwrote;
 	GError *error = NULL;
+	SoupBuffer *chunk;
+	const char *start;
 
 	while (len > io->written) {
 		status = soup_socket_write (io->sock,
@@ -360,7 +367,20 @@ write_data (SoupMessage *msg, const char *data, guint len)
 			return FALSE;
 
 		case SOUP_SOCKET_OK:
+			start = data + io->written;
 			io->written += nwrote;
+
+			if (body) {
+				if (io->write_length)
+					io->write_length -= nwrote;
+
+				chunk = soup_buffer_new (SOUP_MEMORY_TEMPORARY,
+							 start, nwrote);
+				SOUP_MESSAGE_IO_PREPARE_FOR_CALLBACK;
+				soup_message_wrote_body_data (msg, chunk);
+				soup_buffer_free (chunk);
+				SOUP_MESSAGE_IO_RETURN_VAL_IF_CANCELLED_OR_PAUSED (FALSE);
+			}
 			break;
 		}
 	}
@@ -427,11 +447,19 @@ io_write (SoupSocket *sock, SoupMessage *msg)
 			}
 		}
 
-		if (!write_data (msg, io->write_buf->str, io->write_buf->len))
+		if (!write_data (msg, io->write_buf->str,
+				 io->write_buf->len, FALSE))
 			return;
 
 		g_string_truncate (io->write_buf, 0);
 
+		if (io->write_encoding != SOUP_ENCODING_CHUNKED) {
+			SoupMessageHeaders *hdrs =
+				(io->mode == SOUP_MESSAGE_IO_CLIENT) ?
+				msg->request_headers : msg->response_headers;
+			io->write_length = soup_message_headers_get_content_length (hdrs);
+		}
+
 		if (io->mode == SOUP_MESSAGE_IO_SERVER &&
 		    SOUP_STATUS_IS_INFORMATIONAL (msg->status_code)) {
 			if (msg->status_code == SOUP_STATUS_CONTINUE) {
@@ -488,22 +516,47 @@ io_write (SoupSocket *sock, SoupMessage *msg)
 
 
 	case SOUP_MESSAGE_IO_STATE_BODY:
-		if (!io->write_chunk)
-			io->write_chunk = soup_message_body_flatten (io->write_body);
+		if (!io->write_length) {
+			io->write_state = SOUP_MESSAGE_IO_STATE_FINISHING;
+
+			SOUP_MESSAGE_IO_PREPARE_FOR_CALLBACK;
+			soup_message_wrote_body (msg);
+			SOUP_MESSAGE_IO_RETURN_IF_CANCELLED_OR_PAUSED;
+			break;
+		}
+
+		if (!io->write_chunk) {
+			io->write_chunk = soup_message_body_get_chunk (io->write_body, io->write_body_offset);
+			if (!io->write_chunk) {
+				soup_message_io_pause (msg);
+				return;
+			}
+			if (io->write_chunk->length > io->write_length) {
+				/* App is trying to write more than it
+				 * claimed it would; we have to truncate.
+				 */
+				SoupBuffer *truncated =
+					soup_buffer_new_subbuffer (io->write_chunk,
+								   0, io->write_length);
+				soup_buffer_free (io->write_chunk);
+				io->write_chunk = truncated;
+			}
+		}
+
 		if (!write_data (msg, io->write_chunk->data,
-				 io->write_chunk->length))
+				 io->write_chunk->length, TRUE))
 			return;
+
+		soup_message_body_wrote_chunk (io->write_body, io->write_chunk);
 		soup_buffer_free (io->write_chunk);
+		io->write_body_offset += io->write_chunk->length;
 		io->write_chunk = NULL;
 
-		io->write_state = SOUP_MESSAGE_IO_STATE_FINISHING;
-
 		SOUP_MESSAGE_IO_PREPARE_FOR_CALLBACK;
-		soup_message_wrote_body (msg);
+		soup_message_wrote_chunk (msg);
 		SOUP_MESSAGE_IO_RETURN_IF_CANCELLED_OR_PAUSED;
 		break;
 
-
 	case SOUP_MESSAGE_IO_STATE_CHUNK_SIZE:
 		if (!io->write_chunk) {
 			io->write_chunk = soup_message_body_get_chunk (io->write_body, io->write_body_offset);
@@ -516,7 +569,8 @@ io_write (SoupSocket *sock, SoupMessage *msg)
 			io->write_body_offset += io->write_chunk->length;
 		}
 
-		if (!write_data (msg, io->write_buf->str, io->write_buf->len))
+		if (!write_data (msg, io->write_buf->str,
+				 io->write_buf->len, FALSE))
 			return;
 
 		g_string_truncate (io->write_buf, 0);
@@ -533,9 +587,10 @@ io_write (SoupSocket *sock, SoupMessage *msg)
 
 	case SOUP_MESSAGE_IO_STATE_CHUNK:
 		if (!write_data (msg, io->write_chunk->data,
-				 io->write_chunk->length))
+				 io->write_chunk->length, TRUE))
 			return;
 
+		soup_message_body_wrote_chunk (io->write_body, io->write_chunk);
 		soup_buffer_free (io->write_chunk);
 		io->write_chunk = NULL;
 
@@ -550,7 +605,7 @@ io_write (SoupSocket *sock, SoupMessage *msg)
 
 	case SOUP_MESSAGE_IO_STATE_CHUNK_END:
 		if (!write_data (msg, SOUP_MESSAGE_IO_EOL,
-				 SOUP_MESSAGE_IO_EOL_LEN))
+				 SOUP_MESSAGE_IO_EOL_LEN, FALSE))
 			return;
 
 		io->write_state = SOUP_MESSAGE_IO_STATE_CHUNK_SIZE;
@@ -559,7 +614,7 @@ io_write (SoupSocket *sock, SoupMessage *msg)
 
 	case SOUP_MESSAGE_IO_STATE_TRAILERS:
 		if (!write_data (msg, SOUP_MESSAGE_IO_EOL,
-				 SOUP_MESSAGE_IO_EOL_LEN))
+				 SOUP_MESSAGE_IO_EOL_LEN, FALSE))
 			return;
 
 		io->write_state = SOUP_MESSAGE_IO_STATE_FINISHING;
@@ -865,6 +920,11 @@ soup_message_io_pause (SoupMessage *msg)
 		g_signal_handler_disconnect (io->sock, io->read_tag);
 		io->read_tag = 0;
 	}
+
+	if (io->unpause_source) {
+		g_source_destroy (io->unpause_source);
+		io->unpause_source = NULL;
+	}
 }
 
 static gboolean
@@ -874,6 +934,7 @@ io_unpause_internal (gpointer msg)
 	SoupMessageIOData *io = priv->io_data;
 
 	g_return_val_if_fail (io != NULL, FALSE);
+	io->unpause_source = NULL;
 
 	if (io->write_tag || io->read_tag)
 		return FALSE;
@@ -910,9 +971,12 @@ soup_message_io_unpause (SoupMessage *msg)
 		      SOUP_SOCKET_FLAG_NONBLOCKING, &non_blocking,
 		      SOUP_SOCKET_ASYNC_CONTEXT, &async_context,
 		      NULL);
-	if (non_blocking)
-		soup_add_idle (async_context, io_unpause_internal, msg);
-	else
+	if (non_blocking) {
+		if (!io->unpause_source) {
+			io->unpause_source = soup_add_completion (
+				async_context, io_unpause_internal, msg);
+		}
+	} else
 		io_unpause_internal (msg);
 	if (async_context)
 		g_main_context_unref (async_context);
